/*
***************************************
*** Mauro Graziani Minimal Software ***
*** Conversion program from Mac raw ***
*** 24 bit audiofile to 16 bit wav  ***
***************************************

Copyright (c) Mauro Graziani 2006
Site:  www.maurograziani.org
eMail: mauro@maurograziani.org
This program is released under GPL license

*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define	VERSION	"1.0.0"

//------------------------------------------------------------------------------
// header of wav file
typedef struct{
  char rID[4];            // 'RIFF'
  long int rLen;
  char wID[4];            // 'WAVE'
  char fId[4];            // 'fmt '
  long int pcm_header_len;   // varies...
  short int wFormatTag;
  short int nChannels;      // 1,2 for stereo data is (l,r) pairs
  long int nSamplesPerSec;
  long int nAvgBytesPerSec;
  short int nBlockAlign;
  short int nBitsPerSample;
}   WAV_HDR;

// header of wav file
typedef struct{
  char dId[4];            // 'data' or 'fact'
  long int dLen;
  //   unsigned char *data;
}   CHUNK_HDR;

int iBPS = 16, iSR = 44100, iChan = 1;

int WriteWav(short *iSound, int iTotSamples, char *sFile);

//------------------------------------------------------------------------------

int main (int argc, char *argv[])
{
  int i, sz, b, bVerbose = 1;
  short *sam, v;
  int iTotSamples;
  FILE *h;
  unsigned char c[3];
  struct stat st;

  // extern declarations for getopt
  extern char *optarg;
  extern int optind, opterr, optopt;

  if (argc < 2) {
    if (bVerbose) fprintf(stderr, "Usage: %s [options] filename\n", argv[0]);
    exit(255);
  }

  // get command line options
  while ((b = getopt (argc, argv, "shV")) != -1) {

    switch (b) {

    case 's':			// silent
      bVerbose = 0;
      break;

    case 'h':			// help
      fprintf(stdout, "Usage: %s [options] filename\n", argv[0]);
      puts("Options:");
      puts("-s:\tsilent, don't print any msg");
      puts("-V:print version number, some info and exit");
      puts("-h:\tthis help");
      exit(0);
      break;

    case 'V':			// version
      puts("");
      puts("***************************************");
      puts("*** Mauro Graziani Minimal Software ***");
      puts("*** Conversion program from Mac raw ***");
      puts("*** 24 bit audiofile to 16 bit wav  ***");
      puts("***************************************");
      fprintf(stdout, "Version: %s\n", VERSION);
      puts("Copyright (c) Mauro Graziani 2006");
      puts("Site:  www.maurograziani.org");
      puts("eMail: mauro@maurograziani.org");
      puts("This program is released under GPL licence");
      exit(0);
      break;
    }
  }

  if (bVerbose) {
    puts("");
    puts("***************************************");
    puts("*** Mauro Graziani Minimal Software ***");
    puts("*** Conversion program from Mac raw ***");
    puts("*** 24 bit audiofile to 16 bit wav  ***");
    puts("***************************************");
    puts("");
    fprintf(stdout, "Convert %s to %s.16bit.wav\n", argv[optind], argv[optind]);
  }

  // get file dim to compute sample number
  if (stat(argv[optind], &st)) {
    if (bVerbose) perror("*** Error");
    exit(255);
  }
  sz = (int)st.st_size;
  iTotSamples = sz / 3;	// 24 bit samples
  sam = (short *)calloc(iTotSamples, sizeof(short));
  if (!sam) {
    if (bVerbose) fprintf(stderr, "Err alloc mem %d bytes\n", iTotSamples * sizeof(short));
    exit(255);
  }

  h = fopen(argv[optind], "rb");
  if (!h) {
    if (bVerbose) perror("*** Error");
    free((void *)sam);
    exit(255);
  }

  for (i = 0; i < iTotSamples; i++) {
    // read 24 bit mac
    if (fread(&c, sizeof(unsigned char), 3, h) != 3) {
      if (bVerbose) perror("*** Error");
      fclose(h);
      free((void *)sam);
      exit(255);
    }

    // write 2nd then 1st, skip 3rd
    v = c[0];
    v = (v << 8) + c[1];
    *(sam + i) = v;
  }

  if (WriteWav(sam, iTotSamples, argv[optind])) {
    if (bVerbose) perror("*** Error");
    fclose(h);
    free((void *)sam);
    exit(255);
  }

  fclose(h);
  free((void *)sam);
  if (bVerbose)
    puts("");

}
//---------------------------------------------------------------------------
int WriteWav(short *iSound, int iTotSamples, char *sFile)
{
  int i, wbuff_len = iTotSamples * 2 * iChan;
  char obuff[255];
  FILE *w;
  WAV_HDR *wav;
  CHUNK_HDR *chk;

  // allocate wav header
  wav = (WAV_HDR *)malloc(sizeof(WAV_HDR));
  chk = (CHUNK_HDR *)malloc(sizeof(CHUNK_HDR));

  // setup wav header
  sprintf(obuff,"RIFF");
  for(i = 0; i < 4; i++)
    wav->rID[i] = obuff[i];

  sprintf(obuff,"WAVE");
  for(i = 0; i < 4; i++)
    wav->wID[i] = obuff[i];

  sprintf(obuff,"fmt ");
  for(i = 0; i < 4; i++)
    wav->fId[i] = obuff[i];

  wav->nBitsPerSample = iBPS;
  wav->nSamplesPerSec = iSR;
  wav->nChannels = iChan;
  wav->nAvgBytesPerSec = wav->nSamplesPerSec;
  wav->nAvgBytesPerSec *= iBPS / 8;
  wav->nAvgBytesPerSec *= wav->nChannels;

  wav->pcm_header_len = 16;
  wav->wFormatTag = 1;
  wav->rLen = sizeof(WAV_HDR) + sizeof(CHUNK_HDR) + wbuff_len;
  wav->nBlockAlign = iChan * iBPS / 8;

  // setup chunk header
  sprintf(obuff,"data");
  for(i = 0; i < 4; i++)
    chk->dId[i] = obuff[i];

  chk->dLen = wbuff_len;


  strcpy(obuff, sFile);
  strcat(obuff, ".16bit.wav");
  w = fopen(obuff, "wb");
  if (!w)
    return 255;


  if (fwrite(wav, sizeof(WAV_HDR), 1, w) != 1) {
    fclose(w);
    return 255;
  }
  if (fwrite(chk, sizeof(CHUNK_HDR), 1, w) != 1) {
    fclose(w);
    return 255;
  }
  if (fwrite(iSound, sizeof(short), iTotSamples, w) != iTotSamples) {
    fclose(w);
    return 255;
  }

  fclose(w);

  return 0;
}
